@@ -18,7 +18,9 @@ from PySide6.QtCore import (
     QMutexLocker,
     QAbstractTableModel,
     QModelIndex,
-    QPersistentModelIndex
+    QPersistentModelIndex,
+    QCommandLineParser,
+    QElapsedTimer
 )
 from PySide6.QtGui import (
     QFocusEvent,
@@ -47,7 +49,8 @@ from PySide6.QtWidgets import (
     QGroupBox,
     QHeaderView,
     QTableView,
-    QFileDialog
+    QFileDialog,
+    QProgressBar
 )
 from candle.candle_api import (
     CandleDevice,
@@ -61,6 +64,7 @@ from candle.candle_api import (
     GSCANFeature,
     DLC2LEN
 )
+from candle import __version__
 
 
 class CandleManagerState(Enum):
@@ -76,6 +80,7 @@ class CandleManager(QObject):
 
     stateTransition = Signal(CandleManagerState, CandleManagerState)
     messageReceived = Signal(GSHostFrame)
+    busLoad = Signal(int)
     exceptionOccurred = Signal(str)
     channelInfo = Signal(GSDeviceBTConstExtended)
 
@@ -87,6 +92,17 @@ class CandleManager(QObject):
         self.interface: Optional[CandleInterface] = None
         self.channel: Optional[CandleChannel] = None
 
+        self.arbitration_rate: Optional[float] = None
+        self.payload_rate: Optional[float] = None
+        self.total_transfer_time: float = 0
+        self.transfer_elapsed_timer = QElapsedTimer()
+        self.transfer_elapsed_timer.start()
+        self.transfer_time_mutex = QMutexLocker(QMutex())
+        self.bus_load_calculation_timer = QTimer()
+        self.bus_load_calculation_timer.setInterval(1000)
+        self.bus_load_calculation_timer.timeout.connect(self.calculate_bus_load)
+        self.bus_load_calculation_timer.start()
+
         self.state_transition_mutex = QMutexLocker(QMutex())
 
         self.polling_timer = QTimer()   # Do not set parent, timers cannot be stopped from another thread.
@@ -149,20 +165,34 @@ class CandleManager(QObject):
     def set_bit_timing(self, bit_timing: GSDeviceBitTiming) -> None:
         with self.state_transition_mutex:
             if self.state == CandleManagerState.Configuration:
-                self.channel.set_bit_timing(bit_timing.prop_seg, bit_timing.phase_seg1, bit_timing.phase_seg2, bit_timing.sjw, bit_timing.brp)
+                try:
+                    self.channel.set_bit_timing(bit_timing.prop_seg, bit_timing.phase_seg1, bit_timing.phase_seg2, bit_timing.sjw, bit_timing.brp)
+                except usb.core.USBError as e:
+                    self.handle_exception(str(e))
+                else:
+                    self.arbitration_rate = self.channel.clock_frequency / ((1 + bit_timing.prop_seg + bit_timing.phase_seg1 + bit_timing.phase_seg2) * bit_timing.brp)
 
     @Slot(GSDeviceBitTiming)
     def set_data_bit_timing(self, bit_timing: GSDeviceBitTiming) -> None:
         with self.state_transition_mutex:
             if self.state == CandleManagerState.Configuration:
-                self.channel.set_data_bit_timing(bit_timing.prop_seg, bit_timing.phase_seg1, bit_timing.phase_seg2, bit_timing.sjw, bit_timing.brp)
+                try:
+                    self.channel.set_data_bit_timing(bit_timing.prop_seg, bit_timing.phase_seg1, bit_timing.phase_seg2, bit_timing.sjw, bit_timing.brp)
+                except usb.core.USBError as e:
+                    self.handle_exception(str(e))
+                else:
+                    self.payload_rate = self.channel.clock_frequency / ((1 + bit_timing.prop_seg + bit_timing.phase_seg1 + bit_timing.phase_seg2) * bit_timing.brp)
 
     @Slot(bool, bool, bool, bool, bool, bool)
     def start(self, fd: bool, loopback: bool, listen_only: bool, triple_sample: bool, one_shot: bool, bit_error_reporting: bool) -> None:
         with self.state_transition_mutex:
             if self.state == CandleManagerState.Configuration:
-                self.channel.open(fd, loopback, listen_only, triple_sample, one_shot, bit_error_reporting)
-                self.transition(CandleManagerState.Running)
+                try:
+                    self.channel.open(fd, loopback, listen_only, triple_sample, one_shot, bit_error_reporting)
+                except usb.core.USBError as e:
+                    self.handle_exception(str(e))
+                else:
+                    self.transition(CandleManagerState.Running)
 
     @Slot()
     def stop(self) -> None:
@@ -175,7 +205,10 @@ class CandleManager(QObject):
     def set_termination(self, state: bool) -> None:
         with self.state_transition_mutex:
             if self.state == CandleManagerState.Configuration:
-                self.channel.termination = state
+                try:
+                    self.channel.termination = state
+                except usb.core.USBError as e:
+                    self.handle_exception(str(e))
 
     @Slot(GSHostFrame)
     def send_message(self, frame: GSHostFrame) -> None:
@@ -183,7 +216,7 @@ class CandleManager(QObject):
             if self.state == CandleManagerState.Running:
                 try:
                     self.channel.write(frame)
-                except usb.core.USBError as e:
+                except TimeoutError as e:
                     try:
                         self.channel.close()
                     except usb.core.USBError:
@@ -199,20 +232,79 @@ class CandleManager(QObject):
     def handle_exception(self, error: str) -> None:
         self.interface = None
         self.channel = None
+        self.arbitration_rate = None
+        self.payload_rate = None
         self.transition(CandleManagerState.DeviceSelection)
         self.exceptionOccurred.emit(error)
 
+    def update_history(self, frame) -> None:
+        if frame.header.is_error_frame:
+            # Ignore error frame.
+            return
+
+        if self.arbitration_rate is None:
+            return
+
+        if frame.header.is_fd and frame.header.is_bitrate_switch and self.payload_rate is None:
+            return
+
+        # SOF RTR/RRS IDE r0/FDF ACK EOF IFS
+        arbitration_bits = 16
+
+        # ID
+        if frame.header.is_extended_id:
+            arbitration_bits += 29
+        else:
+            arbitration_bits += 11
+
+        # res BRS
+        if frame.header.is_fd:
+            arbitration_bits += 2
+
+        # payload
+        if frame.header.is_fd:
+            if frame.header.data_length > 16:
+                # ESI DLC SBC 21-bit-CRC
+                payload_bits = 31 + frame.header.data_length * 8
+            else:
+                # ESI DLC SBC 17-bit-CRC
+                payload_bits = 27 + frame.header.data_length * 8
+        else:
+            payload_bits = 20 + frame.header.data_length * 8
+
+        if frame.header.is_fd and frame.header.is_bitrate_switch:
+            # BRS
+            transfer_time = arbitration_bits / self.arbitration_rate + payload_bits / self.payload_rate
+        else:
+            transfer_time = (arbitration_bits + payload_bits) / self.arbitration_rate
+
+        with self.transfer_time_mutex:
+            self.total_transfer_time += transfer_time
+
+    @Slot()
+    def calculate_bus_load(self) -> None:
+        if self.transfer_elapsed_timer.elapsed() == 0:
+            return
+
+        with self.transfer_time_mutex:
+            bus_load = 1e3 * self.total_transfer_time / self.transfer_elapsed_timer.elapsed()
+            self.total_transfer_time = 0
+            self.busLoad.emit(round(100 * bus_load))
+
+        self.transfer_elapsed_timer.restart()
+
     @Slot()
     def polling(self) -> None:
         with self.state_transition_mutex:
             if self.state == CandleManagerState.Running:
                 try:
-                    while True:
-                        frame = self.channel.read(1)
-                        if frame is not None:
+                    elapsed_timer = QElapsedTimer()
+                    elapsed_timer.start()
+                    while elapsed_timer.elapsed() < self.polling_timer.interval():
+                        self.interface.polling(1)
+                        while (frame := self.channel.read()) is not None:
+                            self.update_history(frame)
                             self.messageReceived.emit(frame)
-                except usb.core.USBTimeoutError:
-                    pass
                 except usb.core.USBError as e:
                     if e.errno != 32:   # Ignore pipe error.
                         try:
@@ -276,7 +368,6 @@ class InputPanel(QWidget):
                 line_edit.setEnabled(True)
             else:
                 line_edit.setEnabled(False)
-                line_edit.setText('00')
 
     @Slot()
     def random(self) -> None:
@@ -301,6 +392,7 @@ class InputPanel(QWidget):
 
 class MessageTableModel(QAbstractTableModel):
     rowInserted = Signal(int, int)
+    exportFinished = Signal()
 
     def __init__(self, parent: Optional[QObject] = None) -> None:
         super().__init__(parent)
@@ -316,6 +408,15 @@ class MessageTableModel(QAbstractTableModel):
         self.flush_timer.setInterval(50)
         self.flush_timer.start()
 
+    @Slot(str)
+    def export(self, file_path: str) -> None:
+        with open(file_path, 'w', newline='') as csv_file:
+            csv_writer = csv.writer(csv_file)
+            csv_writer.writerow(self.header)
+            for i in range(self.rowCount()):
+                csv_writer.writerow([self.data(self.index(i, j)) for j in range(self.columnCount())])
+        self.exportFinished.emit()
+
     @Slot(GSHostFrame)
     def handle_message(self, message: GSHostFrame) -> None:
         with self.mutex:
@@ -354,7 +455,7 @@ class MessageTableModel(QAbstractTableModel):
             if column == 0:
                 return str(message.timestamp)
             if column == 1:
-                return f'{message.header.arbitration_id:08X}' if message.header.is_extended_id else f'{message.header.arbitration_id:03X}'
+                return f'0x{message.header.arbitration_id:08X}' if message.header.is_extended_id else f'0x{message.header.arbitration_id:03X}'
             if column == 2:
                 return 'Rx' if message.header.is_rx else 'Tx'
             if column == 3:
@@ -669,6 +770,8 @@ class BitTimingDialog(QDialog):
 
 
 class MainWindow(QWidget):
+    export = Signal(str)
+
     def __init__(self, parent: Optional[QWidget] = None) -> None:
         super().__init__(parent)
         self.setWindowTitle('Candle Viewer')
@@ -789,6 +892,17 @@ class MainWindow(QWidget):
         hbox_layout3.addWidget(self.input_panel)
         hbox_layout3.addLayout(vbox_layout2)
         hbox_layout3.addSpacerItem(QSpacerItem(20, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum))
+        vbox_layout_3 = QVBoxLayout()
+        bus_load_bar = QProgressBar()
+        bus_load_bar.setOrientation(Qt.Vertical)
+        bus_load_bar.setMaximum(100)
+        bus_load_bar.setValue(0)
+        bus_load_bar.setTextVisible(False)
+        bus_load_label = QLabel('0%')
+        vbox_layout_3.addWidget(QLabel('Load'), alignment=Qt.AlignmentFlag.AlignHCenter)
+        vbox_layout_3.addWidget(bus_load_bar, alignment=Qt.AlignmentFlag.AlignHCenter)
+        vbox_layout_3.addWidget(bus_load_label, alignment=Qt.AlignmentFlag.AlignHCenter)
+        hbox_layout3.addLayout(vbox_layout_3)
         vbox_layout.addLayout(hbox_layout1)
         vbox_layout.addLayout(hbox_layout2)
         vbox_layout.addWidget(self.message_viewer, 1)
@@ -840,6 +954,10 @@ class MainWindow(QWidget):
         export_button.clicked.connect(self.handle_export)
         self.termination_checkbox.toggled.connect(self.candle_manager.set_termination)
         self.candle_manager.selectDeviceResult.connect(self.handle_select_device_result)
+        self.export.connect(self.message_model.export)
+        self.message_model.exportFinished.connect(self.handle_export_finished)
+        self.candle_manager.busLoad.connect(bus_load_bar.setValue)
+        self.candle_manager.busLoad.connect(lambda x: bus_load_label.setText(f'{x}%'))
 
         # Start thread and timer.
         self.polling_thread.start()
@@ -887,7 +1005,7 @@ class MainWindow(QWidget):
     @Slot(CandleManagerState)
     def handle_state_transition(self, _from_state: CandleManagerState, to_state: CandleManagerState) -> None:
         if to_state == CandleManagerState.DeviceSelection:
-            self.device_selector.setEnabled(True)
+            self.device_selector.setEnabled(False)
             self.channel_selector.setEnabled(False)
             self.bit_timing_button.setEnabled(False)
             self.start_button.setEnabled(False)
@@ -1038,12 +1156,17 @@ class MainWindow(QWidget):
     @Slot()
     def handle_export(self) -> None:
         file_name = QFileDialog.getSaveFileName(self, filter="CSV (*.csv)")[0]
-        if file_name:
-            with open(file_name, 'w', newline='') as csv_file:
-                csv_writer = csv.writer(csv_file)
-                csv_writer.writerow(self.message_model.header)
-                for i in range(self.message_model.rowCount()):
-                    csv_writer.writerow([self.message_model.data(self.message_model.index(i, j)) for j in range(self.message_model.columnCount())])
+        if not file_name:
+            return
+        if not file_name.endswith('.csv'):
+            file_name += '.csv'
+        self.export.emit(file_name)
+
+    @Slot()
+    def handle_export_finished(self) -> None:
+        message_box = QMessageBox(self)
+        message_box.setText('Export Finished')
+        message_box.open()
 
     def closeEvent(self, event: QCloseEvent):
         self.polling_thread.requestInterruption()
@@ -1058,6 +1181,15 @@ class MainWindow(QWidget):
 def main():
     app = QApplication(sys.argv)
     app.setStyle('Fusion')
+    app.setApplicationName("candle_viewer")
+    app.setApplicationVersion(__version__)
+
+    parser = QCommandLineParser()
+    parser.setApplicationDescription("Candle Viewer")
+    parser.addHelpOption()
+    parser.addVersionOption()
+    parser.process(app)
+
     main_window = MainWindow()
     main_window.show()
     return app.exec()
